const { asyncMapValuesDeep } = require('./util');

const _ = require('lodash');
const md5 = require('md5');
const request = require('request-json');

const createRedisClient = require('../util/redis');
const AsksuiteUtil = require('../asksuite.util')();

const languageMap = {
  pt_br: 'ptBR',
  pt: 'ptBR',
  'pt-br': 'ptBR',
  en: 'enUS',
  es: 'es',
};

module.exports = configuration => {
  const redis = createRedisClient(configuration.redis);

  const getTranslation = async (text, languageFrom, languageTo, customId, options) => {
    if (languageTo === languageFrom) return text;

    const client = request.createClient(configuration.url);

    let path = `translate/${languageTo}/${languageFrom}`;
    if (customId) {
      path += `/${customId}`;
    }

    const optionsKeys = _.keys(options);

    if (!_.isEmpty(optionsKeys)) {
      const queryString = _.map(optionsKeys, key => `${key}=${_.get(options, key)}`).join('&');
      path += `?${queryString}`;
    }

    const translateResponse = await client.post(path, [text]);
    return translateResponse.body[0];
  };

  const getCachedTranslation = async (hash, languageFormatIso, languageFormatted, customId) => {
    const customTranslation = await redis.getValue(`${customId}:${hash}`);
    const globalTranslation = await redis.getValue(hash);

    const path = `translations.${languageFormatIso}`;
    const languageProperty = _.get(languageMap, languageFormatted);
    return (
      _.get(customTranslation, path) ||
      _.get(globalTranslation, path) ||
      _.get(customTranslation, languageProperty) ||
      _.get(globalTranslation, languageProperty)
    );
  };

  const normalizeLanguage = (lang, languages) => {
    if (languages) {
      const l = AsksuiteUtil.resolveLanguage(lang, languages, 'formatIso');
      if (l) {
        return l;
      }
    }
    return lang
      .toLowerCase()
      .replace('-', '_')
      .split('_')[0];
  };

  const notNeedsTranslate = (
    text,
    attribute,
    removedAttributes,
    languageFrom,
    languageTo,
    languages,
  ) => {
    const regexRemovedAttributes = _.filter(removedAttributes, _.isRegExp);
    const notNeedsByAllRegexRemovedAttributes = _.some(regexRemovedAttributes, regex =>
      regex.test(text),
    );

    if (notNeedsByAllRegexRemovedAttributes) {
      return notNeedsByAllRegexRemovedAttributes;
    }

    const notNeeds =
      _.includes(removedAttributes, attribute) || !_.isString(text) || _.isEmpty(text);
    if (notNeeds) {
      return notNeeds;
    }

    return normalizeLanguage(languageFrom, languages) === normalizeLanguage(languageTo, languages);
  };

  const translate = async (
    text,
    attribute,
    removedAttributes,
    languageFrom,
    languageTo,
    customId,
    options,
    languages,
  ) => {
    if (notNeedsTranslate(text, attribute, removedAttributes, languageFrom, languageTo, languages))
      return text;

    const hash = md5(text);
    const languageFormatIso = normalizeLanguage(languageTo, languages);
    const languageFormatted = normalizeLanguage(languageTo, null);

    const cachedTranslation = await getCachedTranslation(
      hash,
      languageFormatIso,
      languageFormatted,
      customId,
    );

    if (cachedTranslation) {
      return cachedTranslation || text;
    }

    const translation = await getTranslation(text, languageFrom, languageTo, customId, options);

    return translation || text;
  };

  const translateObject = (
    object,
    languageFrom,
    languageTo,
    removedAttributes,
    customId,
    options,
    languages,
  ) => {
    const contextTranslate = (text, attribute) =>
      translate(
        text,
        attribute,
        removedAttributes,
        languageFrom,
        languageTo,
        customId,
        options,
        languages,
      );
    return asyncMapValuesDeep(object, contextTranslate);
  };

  return translateObject;
};
